# This file includes tools to modify the code and make the implementation easier
def expand_sources(source_path):
    # Get all files contained in a source path (recursive)
    from os import walk
    from os.path import join
    files = [join(dp, f) for dp, dn, fn in walk(source_path) for f in fn]
    file_paths = {}
    for f in files:
        name = f.split("/")[-1]
        file_paths[name] = f
    return file_paths


def match_module_and_filename(module, filenames):
    # Gets the filepath of the corresponding module
    import re
    pattern = "%s\.[fFh90]" % module
    matches = [fn for fn in filenames if re.match(pattern, fn)]
    assert len(matches) == 1, "Multiple file match for module %s" % module
    return matches[0]


def find_declaration_line(variable, filepath, vault):
    # Finds line at which the variable is declared
    import re
    from UtilsRPE.Functions import in_which_block
    import warnings
    # Declaration pattern
    # TODO: This might have issues with split lines
    pattern = "::.*\W%s\W" % variable.name
    with open(filepath, "r") as f:
        lines = [l for l in f]

        current_block = vault[variable.routine.module.name].main()

        for _index, _line in enumerate(lines):
            if re.search(pattern, _line):
                if current_block == variable.routine:
                    return _index
            else:
                current_block = in_which_block(_line, current_block, vault=vault)
    warnings.warn("Declaration of variable %s not found in module %s" % (variable.name, variable.routine.module.name))
    return None


def replace_declaration(variable, file_paths, vault, target_precision="dp"):
    # This function will find the file where a variable is declared and will replace the declaration
    # with a new one using the new precision spec.
    files = file_paths.keys()
    module = variable.routine.module.name
    filepath = file_paths[match_module_and_filename(module, files)]
    declaration_line_index = find_declaration_line(variable, filepath, vault)
    if declaration_line_index is not None:
        with open(filepath, "r") as f:
            data = [l for l in f]
            declaration_line = data[declaration_line_index]
            new_line = replace_real_declaration(variable, declaration_line, target_precision=target_precision)
            data[declaration_line_index] = new_line
            updated_text = "".join(data)
        with open(filepath, "w") as output:
            output.write(updated_text)


def replace_real_declaration(variable, declaration_line, target_precision="dp"):
    # This function takes a declaration line (string) and updates it with the new type
    # If the declaration contains more variables it keeps the other variables with the original attributes.
    from AutoRPE.UtilsRPE.Functions import remove_comments, split_declaration
    import re
    import warnings
    declaration_line_without_comments = remove_comments(declaration_line).strip()

    attributes, arguments = split_declaration(declaration_line_without_comments)

    # Real statement pattern
    pattern = "real *\(([^\)]*)\)"

    if len(arguments) == 1:
        if declaration_line_without_comments[-1] == "&":
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
        else:
            m = re.search(pattern, declaration_line_without_comments, re.I)
            if m:
                real_statement = m.group(0)
                precision_specification = m.group(1)
                new_real_statement = real_statement.replace(precision_specification, target_precision)
                new_declaration_line_without_comments = \
                    declaration_line_without_comments.replace(real_statement, new_real_statement)
                new_declaration_line = \
                    declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
            else:
                warnings.warn("Problems finding real statement:\n%s" % declaration_line)
                new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                       declaration_line
        return new_declaration_line
    else:
        if variable.name not in arguments:
            raise AssertionError("Variable %s not found in list of arguments" % variable.name, attributes)
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
            leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line


def change_code(list_filename, source_path, vault_path):
    # The purpose of this function is to modify the declarations of several variables in a source-code.
    # As inputs it takes the path to the file with the list of variables, the path to the sources that will be modified
    # and the path to the vault database containing information about the code.
    #
    # This function takes the following inputs:
    # - list_filename: Path to a file that contains a list of variables that will be modified in the code
    # - source_path: Path to a folder that contains the source files of the code corresponding to the vault.
    # - vault_path: Path to a vault database.
    from UtilsRPE.Functions import load_vault, load_variable_list

    vault = load_vault(vault_path)
    variable_ids = load_variable_list(list_filename)
    file_paths = expand_sources(source_path=source_path)

    variables = [vault.get_variable_by_id(_id) for _id in variable_ids]

    for variable in variables:
        replace_declaration(variable, file_paths, vault)

    pass


def update_vault_with_precision_from_source(source_path, vault_path, lookup_table=None):
    # The purpose of this function is to read a specific repository of sources and update the precision of the vault
    # variables accordingly.
    # It can be useful to reproduce the results of a specific code implementation using the emulator.
    #
    # This function takes the following inputs:
    # - source_path: Path to a folder that contains the source files of the code corresponding to the vault.
    # - vault_path: Path to a vault database.
    # - lookup_table: dictionary that contains the equivalence between the real types that appear in the code and
    #                 its number of significant bits. i.e.  lookup_table = {"dp": 52, "sp":23}
    # If any lookup table is provided,the function will print a list of the different real type specifications
    # that appear in the code, that should make possible to create the lookup table.
    from UtilsRPE.Functions import load_vault, in_which_block, split_declaration, split_attributes
    from UtilsRPE.Classes import ReadSourceFile
    import re
    file_paths = expand_sources(source_path=source_path)
    vault = load_vault(vault_path)

    pattern = ".*REAL.*:: *([^\W]*)\W*"
    real_pat = ".*REAL.*::.*"
    if not lookup_table:
        used_types = []
    # We will loop through the modules
    for module in vault.modules:
        # For each module in the vault, we will find the file in the sources provided
        module_file = match_module_and_filename(module.name, file_paths)
        module_filepath = file_paths[module_file]
        # Loading the file
        f = ReadSourceFile(module_filepath)

        # We start in the main block
        current_block = vault[module.name].main()

        # Loop through the file lines
        for _index, _line in enumerate(f.lines):
            # Keep track of which block are we currently in
            if current_block is None:
                current_block = in_which_block(_line, current_block, vault=vault)
            else:
                # Look for declaration lines.
                # If the file has been pre-processed in each line we will have a single variable
                m = re.search(pattern, _line, re.I)
                if m:
                    # Find variable name
                    var_name = m.group(1)
                    found = False
                    # Match current variable with corresponding variable in the vault routine
                    for var in current_block.variables:
                        if var.name == var_name:
                            found = True
                            vault_variable = var
                            break
                    # Warning in case variable is not found
                    # It might happen that a variable appearing into the full source code is not contained in the vault
                    # because the vault was build from preprocessed sources in which this variable has not been defined.
                    if not found:
                        pass
                        # print("%s not found" % var_name)
                    else:
                        attributes, arguments = split_declaration(_line)
                        real_dec = split_attributes(attributes)[0]
                        pat = "real\((.*)\)"
                        m = re.match(pat, real_dec, re.I)
                        if m:
                            nm = re.match("real\(.*=(.*)\)", real_dec, re.I)
                            if nm:
                                var_type = nm.group(1)
                            else:
                                var_type = m.group(1)

                        elif re.search("real", real_dec, re.I):
                            pass
                            var_type = "default"
                        else:
                            raise AssertionError("WTF!")
                        if not lookup_table:
                            used_types.append(var_type)
                        else:
                            vault_variable.precision = lookup_table[var_type]



                else:
                    # If the line is not a declaration, chech if we are switching to a different block
                    current_block = in_which_block(_line, current_block, vault=vault)

    if not lookup_table:
        print("You didn't set up the lookup table. "
              "You have to specify the types of the following cases appearing in the code:")
        print(set(used_types))
    return vault


def get_variable_id(vault, variable_name, routine_name=None, module_name=None, ):
    # Small function created to easily find the id of variables that appear in the code
    vs = [v for v in vault.variables if v.name == variable_name]

    if module_name is not None:
        vs = [v for v in vs if v.routine.module.name == module_name]
    if routine_name is not None:
        vs = [v for v in vs if v.routine.name == routine_name]

    if len(vs) == 1:
        print("Variable found! ID: %i" % vs[0].id)
        return vs[0].id
    elif len(vs) > 1:
        print("More than one variable!")
        return None
    else:
        print("Not a single variable found!")
        return None


def expand_vault(vault):
    # The purpose of this function is to assign ids to rpe_variables that didn't have any assigned.
    rpe_vars = [v for v in vault.variables if v.type == "rpe_var"]
    for var in rpe_vars:
        var.mutable = True
    rpe_vars_with_id = [v for v in rpe_vars if v.id is not None]
    rpe_vars_without_id = [v for v in rpe_vars if v.id is None]

    last_id = max([v.id for v in rpe_vars_with_id])

    counter = last_id + 1
    for v in rpe_vars_without_id:
        v.id = counter
        v.mutable = False
        counter += 1


# Few test cases

def case_update_vault_with_precision_from_source():
    source_path = "/home/Earth/otintopr/Workspace/test_codechange/src/OCE"
    vault_path = "/esarchive/scratch/otinto/ExperimentAutoRPE_45days/ConfigurationFiles/vault-NEMO4.0.1.pkl"

    lookup_table = {'WP': 23, 'default': 23, '8': 52, 'dp': 52, 'sp': 23, 'fbsp': 23, 'wp': 23}
    vault = update_vault_with_precision_from_source(source_path=source_path, vault_path=vault_path,
                                                    lookup_table=lookup_table)

    not_found_vars = [v for v in vault.variables if v.id is not None and v.precision is None]
    print(len(not_found_vars))


def case_expand_vault():
    from UtilsRPE.Functions import load_vault
    vault_path = "/esarchive/scratch/otinto/ExperimentAutoRPE_45days/ConfigurationFiles/vault-NEMO4.0.1.pkl"

    vault = load_vault(vault_path)

    expand_vault(vault)

    get_variable_id(vault, "pts", module_name="eosbn2", routine_name="eos_insitu")


def case_expand_plus_update_vault_with_precision():
    from UtilsRPE.Functions import save_variable_list
    source_path = "/home/Earth/otintopr/Workspace/test_codechange/src/OCE"
    vault_path = "/esarchive/scratch/otinto/ExperimentAutoRPE_45days/ConfigurationFiles/vault-NEMO4.0.1.pkl"

    # Define which number of significant bits correspond to each type specification found in the code
    lookup_table = {'WP': 23, 'default': 23, '8': 52, 'dp': 52, 'sp': 23, 'fbsp': 23, 'wp': 23}

    # Read the precision status from source
    vault = update_vault_with_precision_from_source(source_path=source_path, vault_path=vault_path,
                                                    lookup_table=lookup_table)

    # Assign ids to variables that doesn't have one
    expand_vault(vault)

    # Look for the variables that are in double precision in the source and find its ids
    double_precision_variables = [v for v in vault.variables if v.precision == 52]
    variable_ids = [v.id for v in double_precision_variables]

    # Save a list with these variables
    save_variable_list(list_of_variables=variable_ids, filename="/home/Earth/otintopr/test_dp_vars.txt", vault=vault)

    vault.generate_namelist_precisions(filename="/home/Earth/otintopr/test_dp_vars_namelist.txt", exclude_double_precision=False)


if __name__ == "__main__":
    case_expand_plus_update_vault_with_precision()
